#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
This module is a library with some tools and functions for EEG signal analysis.

*Author*: Luis Cabañero Gómez
"""

from . import eeg
from . import helpers
from . import features
from . import preprocessing
from . import auxFunctions
from . import wrapper